# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import os
import stat
import time

import vmdb


class MkpartPlugin(vmdb.Plugin):
    def enable(self):
        self.app.step_runners.add(MkpartStepRunner())


class MkpartStepRunner(vmdb.StepRunnerInterface):
    def get_key_spec(self):
        return {
            "mkpart": str,
            "device": str,
            "start": str,
            "end": str,
            "tag": "",
            "part-tag": "",
            "fs-type": "ext2",
        }

    def run(self, values, settings, state):
        part_type = values["mkpart"]
        device = values["device"]
        start = values["start"]
        end = values["end"]
        tag = values["tag"] or values["part-tag"] or None
        fs_type = values["fs-type"]

        # Enabling this breaks installation to an LVM2 logical volume.
        # device = os.path.realpath(device)

        orig = self.list_partitions(device)
        vmdb.runcmd(
            ["parted", "-s", device, "--", "mkpart", part_type, fs_type, start, end]
        )
        new = self.list_partitions(device)
        diff = self.diff_partitions(orig, new)

        if len(diff) == 0:
            raise ExpectedNewPartition()

        if len(diff) > 1:
            raise UnexpectedNewPartitions(diff)

        state.tags.append(tag)

        # If device is a real block device (e.g, /dev/sdb), the
        # parition we have in diff are also real devices (e.g.,
        # /dev/sdb1), and we should remember those in tags.
        #
        # If, however, device is a disk image file (e.g, foo.img), the
        # partition in diff is not a device file but something like
        # foo.img1. We don't need to remember that in tags. The user
        # will use the kpartx step later to add those partitions into
        # tags.
        if self.is_block_dev(device):
            self.wait_for_file_to_exist(diff[0])
            vmdb.progress("remembering partition {} as {}".format(diff[0], tag))
            state.tags.set_dev(tag, diff[0])

    def is_block_dev(self, filename):
        st = os.lstat(filename)
        return stat.S_ISBLK(st.st_mode)

    def list_partitions(self, device):
        output = vmdb.runcmd(["parted", "-m", device, "print"])
        output = output.decode("UTF-8")
        partitions = [line.split(":")[0] for line in output.splitlines() if ":" in line]
        return [
            word if word.startswith("/") else "{}{}".format(device, word)
            for word in partitions
        ]

    def diff_partitions(self, old, new):
        return [line for line in new if line not in old]

    def wait_for_file_to_exist(self, filename):
        while not os.path.exists(filename):
            time.sleep(1)


class MkpartError(Exception):

    pass


class ExpectedNewPartition(MkpartError):
    def __init__(self):
        super().__init__("Expected a new partition to exist after mkpart")


class UnexpectedNewPartitions(MkpartError):
    def __init__(self, diff):
        super().__init__(
            "Expected only one new partition to exist after mkpart, "
            "but found {}".format(" ".join(diff))
        )



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import os
import tempfile

import vmdb


class AnsiblePlugin(vmdb.Plugin):
    def enable(self):
        self.app.step_runners.add(AnsibleStepRunner())


class AnsibleStepRunner(vmdb.StepRunnerInterface):
    def get_key_spec(self):
        return {"ansible": str, "playbook": str}

    def run(self, values, settings, state):
        tag = values["ansible"]
        playbook = values["playbook"]
        mount_point = state.tags.get_builder_mount_point(tag)
        rootfs_tarball = settings["rootfs-tarball"]

        state.ansible_inventory = self.create_inventory(mount_point)
        vmdb.progress(
            "Created {} for Ansible inventory".format(state.ansible_inventory)
        )

        vars_filename = self.create_vars(rootfs_tarball)
        vmdb.progress("Created {} for Ansible variables".format(vars_filename))

        env = dict(os.environ)
        env["ANSIBLE_NOCOWS"] = "1"
        vmdb.runcmd(
            [
                "ansible-playbook",
                "-c",
                "chroot",
                "-i",
                state.ansible_inventory,
                "-e",
                "@{}".format(vars_filename),
                playbook,
            ],
            env=env,
        )

    def teardown(self, values, settings, state):
        if hasattr(state, "ansible_inventory"):
            vmdb.progress("Removing {}".format(state.ansible_inventory))
            os.remove(state.ansible_inventory)

    def create_inventory(self, chroot):
        fd, filename = tempfile.mkstemp()
        os.write(fd, "[image]\n{}\n".format(chroot).encode())
        os.close(fd)
        return filename

    def create_vars(self, tarball):
        fd, filename = tempfile.mkstemp()
        os.write(fd, 'rootfs_tarball: "{}"\n'.format(tarball).encode())
        os.close(fd)
        return filename



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import logging
import os
import tempfile

import vmdb


class MountPlugin(vmdb.Plugin):
    def enable(self):
        self.app.step_runners.add(MountStepRunner())


class MountStepRunner(vmdb.StepRunnerInterface):
    def get_key_spec(self):
        return {"mount": str, "dirname": "", "mount-on": ""}

    def run(self, values, settings, state):
        self.mount_rootfs(values, settings, state)

    def teardown(self, values, settings, state):
        self.unmount_rootfs(values, settings, state)

    def mount_rootfs(self, values, settings, state):
        tag = values["mount"]
        dirname = values["dirname"] or None
        mount_on = values["mount-on"] or None

        device = state.tags.get_dev(tag)

        if dirname:
            if not mount_on:
                raise Exception("no mount-on tag given")

            if not state.tags.has_tag(mount_on):
                raise Exception("cannot find tag {}".format(mount_on))

            mount_point = os.path.join(
                state.tags.get_builder_mount_point(mount_on), "./" + dirname
            )

            if not os.path.exists(mount_point):
                os.makedirs(mount_point)
        else:
            dirname = "/"
            mount_point = tempfile.mkdtemp()

        vmdb.runcmd(["mount", device, mount_point])
        state.tags.set_builder_mount_point(tag, mount_point, cached=True)
        state.tags.set_target_mount_point(tag, dirname)

        return mount_point

    def unmount_rootfs(self, values, settings, state):
        tag = values["mount"]
        mount_point = state.tags.get_builder_mount_point(tag)
        if mount_point is None:
            return

        try:
            vmdb.unmount(mount_point)
        except vmdb.NotMounted as e:
            logging.warning(str(e))

        if not values["mount-on"]:
            os.rmdir(mount_point)



# Copyright 2019 Antonio Terceiro
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=

import os

import vmdb


class FstabPlugin(vmdb.Plugin):
    def enable(self):
        self.app.step_runners.add(FstabStepRunner())


class FstabStepRunner(vmdb.StepRunnerInterface):
    def get_key_spec(self):
        return {"fstab": str}

    def run(self, values, setting, state):
        tag = values["fstab"]
        chroot = state.tags.get_builder_mount_point(tag)

        filesystems = []

        for tag in state.tags.get_tags():
            device = state.tags.get_dev(tag)
            mount_point = state.tags.get_target_mount_point(tag)
            if mount_point is not None:
                fstype = state.tags.get_fstype(tag)
                output = vmdb.runcmd(
                    ["blkid", "-c", "/dev/null", "-o", "value", "-s", "UUID", device]
                )
                if output:
                    uuid = output.decode().strip()
                    filesystems.append(
                        {"uuid": uuid, "mount_point": mount_point, "fstype": fstype}
                    )
                else:
                    raise Exception(
                        "Unknown UUID for device {} (to be mounted on {})".format(
                            device, mount_point
                        )
                    )

        fstab_path = os.path.join(chroot, "etc/fstab")
        line = "UUID={uuid} {mount_point} {fstype} errors=remount-ro 0 1\n"
        with open(fstab_path, "w") as fstab:
            for entry in filesystems:
                fstab.write(line.format(**entry))



# Copyright 2018  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import logging
import os
import tempfile

import vmdb


class LuksPlugin(vmdb.Plugin):
    def enable(self):
        self.app.step_runners.add(CryptsetupStepRunner())


class CryptsetupStepRunner(vmdb.StepRunnerInterface):
    def get_key_spec(self):
        return {"cryptsetup": str, "tag": str, "key-file": "", "key-cmd": ""}

    def run(self, values, settings, state):
        underlying = values["cryptsetup"]
        crypt_name = values["tag"]

        if not isinstance(underlying, str):
            raise vmdb.NotString("cryptsetup", underlying)
        if not isinstance(crypt_name, str):
            raise vmdb.NotString("cryptsetup: tag", crypt_name)

        state.tmp_key_file = None
        key_file = values["key-file"] or None
        key_cmd = values["key-cmd"] or None
        if key_file is None and key_cmd is None:
            raise Exception("cryptsetup step MUST define one of key-file or key-cmd")

        if key_file is None:
            output = vmdb.runcmd(["sh", "-ec", key_cmd])
            output = output.decode("UTF-8")
            key = output.splitlines()[0]
            fd, key_file = tempfile.mkstemp()
            state.tmp_key_file = key_file
            os.close(fd)
            open(key_file, "w").write(key)

        dev = state.tags.get_dev(underlying)
        if dev is None:
            for t in state.tags.get_tags():
                logging.debug(
                    "tag %r dev %r mp %r",
                    t,
                    state.tags.get_dev(t),
                    state.tags.get_builder_mount_point(t),
                )
            assert 0

        vmdb.runcmd(["cryptsetup", "-q", "luksFormat", dev, key_file])
        vmdb.runcmd(
            [
                "cryptsetup",
                "open",
                "--type",
                "luks",
                "--key-file",
                key_file,
                dev,
                crypt_name,
            ]
        )

        crypt_dev = "/dev/mapper/{}".format(crypt_name)
        assert os.path.exists(crypt_dev)
        state.tags.append(crypt_name)
        state.tags.set_dev(crypt_name, crypt_dev)

    def teardown(self, values, settings, state):
        x = state.tmp_key_file
        if x is not None and os.path.exists(x):
            os.remove(x)

        crypt_name = values["tag"]

        crypt_dev = "/dev/mapper/{}".format(crypt_name)
        vmdb.runcmd(["cryptsetup", "close", crypt_dev])



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import vmdb


class KpartxPlugin(vmdb.Plugin):
    def enable(self):
        self.app.step_runners.add(KpartxStepRunner())


class KpartxStepRunner(vmdb.StepRunnerInterface):
    def get_key_spec(self):
        return {"kpartx": str}

    def run(self, values, settings, state):
        device = values["kpartx"]
        tags = state.tags.get_tags()
        devs = self.kpartx(device)
        for tag, dev in zip(tags, devs):
            vmdb.progress("remembering {} as {}".format(dev, tag))
            state.tags.set_dev(tag, dev)

    def kpartx(self, device):
        output = vmdb.runcmd(["kpartx", "-asv", device]).decode("UTF-8")
        for line in output.splitlines():
            words = line.split()
            if words[0] == "add":
                name = words[2]
                yield "/dev/mapper/{}".format(name)

    def teardown(self, values, settings, state):
        device = values["kpartx"]
        vmdb.runcmd(["kpartx", "-dsv", device])



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import logging
import os

import vmdb


class VirtualFilesystemMountPlugin(vmdb.Plugin):
    def enable(self):
        self.app.step_runners.add(VirtualFilesystemMountStepRunner())


class VirtualFilesystemMountStepRunner(vmdb.StepRunnerInterface):

    virtuals = [
        ["none", "/proc", "proc"],
        ["none", "/dev", "devtmpfs"],
        ["none", "/dev/pts", "devpts"],
        ["none", "/dev/shm", "tmpfs"],
        ["none", "/run", "tmpfs"],
        ["none", "/run/lock", "tmpfs"],
        ["none", "/sys", "sysfs"],
    ]

    def get_key_spec(self):
        return {"virtual-filesystems": str}

    def run(self, values, settings, state):
        fstag = values["virtual-filesystems"]
        mount_point = state.tags.get_builder_mount_point(fstag)
        self.mount_virtuals(mount_point, state)

    def teardown(self, values, settings, state):
        self.unmount_virtuals(state)

    def mount_virtuals(self, rootfs, state):
        if not hasattr(state, "virtuals"):
            state.virtuals = []

        for device, mount_point, fstype in self.virtuals:
            path = os.path.join(rootfs, "./" + mount_point)
            if not os.path.exists(path):
                os.mkdir(path)
            vmdb.runcmd(["mount", "-t", fstype, device, path])
            state.virtuals.append(path)
        logging.debug("mounted virtuals: %r", state.virtuals)

    def unmount_virtuals(self, state):
        logging.debug("unmounting virtuals: %r", state.virtuals)
        for mount_point in reversed(state.virtuals):
            try:
                vmdb.unmount(mount_point)
            except vmdb.NotMounted as e:
                logging.warning(str(e))



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import vmdb
import subprocess


class DebootstrapPlugin(vmdb.Plugin):
    def enable(self):
        self.app.step_runners.add(DebootstrapStepRunner())


class DebootstrapStepRunner(vmdb.StepRunnerInterface):
    def get_key_spec(self):
        return {
            "debootstrap": str,
            "target": str,
            "mirror": str,
            "arch": "",
            "keyring": "",
            "variant": "-",
            "components": ["main"],
        }

    def run(self, values, settings, state):
        suite = values["debootstrap"]
        tag = values["target"]
        target = state.tags.get_builder_mount_point(tag)
        mirror = values["mirror"]
        keyring = values["keyring"] or None
        arch = (
            values["arch"]
            or subprocess.check_output(["dpkg", "--print-architecture"]).strip()
        )
        variant = values["variant"]
        components = values["components"]

        if not (suite and tag and target and mirror):
            raise Exception("missing arg for debootstrap step")
        if keyring:
            vmdb.runcmd(
                [
                    "debootstrap",
                    "--keyring",
                    keyring,
                    "--arch",
                    arch,
                    "--variant",
                    variant,
                    "--components",
                    ",".join(components),
                    suite,
                    target,
                    mirror,
                ]
            )
        else:
            vmdb.runcmd(
                [
                    "debootstrap",
                    "--arch",
                    arch,
                    "--variant",
                    variant,
                    "--components",
                    ",".join(components),
                    suite,
                    target,
                    mirror,
                ]
            )

    def run_even_if_skipped(self, values, settings, state):
        tag = values["target"]
        target = state.tags.get_builder_mount_point(tag)
        vmdb.runcmd_chroot(target, ["apt-get", "update"])



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import logging
import os
import subprocess
import sys


_verbose = False


def set_verbose_progress(verbose):
    global _verbose
    _verbose = verbose


def error(msg):
    logging.error(msg, exc_info=True)
    sys.stderr.write("ERROR: {}\n".format(msg))
    sys.stderr.flush()


def progress(msg):
    logging.info(msg)
    if _verbose:
        sys.stdout.write("{}\n".format(msg))
        sys.stdout.flush()


def runcmd(argv, **kwargs):
    progress("Exec: %r" % (argv,))
    env = kwargs.get("env", os.environ.copy())
    env["LC_ALL"] = "C"
    kwargs["env"] = env
    kwargs["stdout"] = kwargs.get("stdout", subprocess.PIPE)
    kwargs["stderr"] = kwargs.get("stderr", subprocess.PIPE)
    p = subprocess.Popen(argv, **kwargs)
    out, err = p.communicate()
    logging.debug("STDOUT: %s", out.decode("UTF8"))
    logging.debug("STDERR: %s", err.decode("UTF8"))
    if p.returncode != 0:
        raise RuncmdError("Program failed: {}".format(p.returncode))
    return out


def runcmd_chroot(chroot, argv, *argvs, **kwargs):
    full_argv = ["chroot", chroot] + argv
    return runcmd(full_argv, *argvs, **kwargs)


def _procdir(chroot):
    proc = os.path.join(chroot, "proc")
    if not os.path.exists(proc):
        os.mkdir(proc, mode=0o755)
    return proc


def _log_stdout(data):
    logging.debug("STDOUT: %r", data)
    return data


def _log_stderr(data):
    logging.debug("STDERR: %r", data)
    return data


class RuncmdError(Exception):
    pass



# Copyright 2018  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


class Tags:
    def __init__(self):
        self._tags = {}
        self._tagnames = []

    def __repr__(self):  # pragma: no cover
        return repr({"tags": self._tags, "tagnames": self._tagnames})

    def get_tags(self):
        return list(self._tags.keys())

    def has_tag(self, tag):
        return tag in self._tags

    def get_dev(self, tag):
        item = self._get(tag)
        return item["dev"]

    def get_builder_mount_point(self, tag):
        item = self._get(tag)
        return item["builder_mount_point"]

    def get_fstype(self, tag):
        item = self._get(tag)
        return item["fstype"]

    def get_target_mount_point(self, tag):
        item = self._get(tag)
        return item["target_mount_point"]

    def is_cached(self, tag):
        item = self._get(tag)
        return item.get("cached", False)

    def append(self, tag):
        if tag in self._tags:
            raise TagInUse(tag)
        self._tagnames.append(tag)
        self._tags[tag] = {
            "dev": None,
            "builder_mount_point": None,
            "fstype": None,
            "target_mount_point": None,
        }

    def set_dev(self, tag, dev):
        item = self._get(tag)
        if item["dev"] is not None:
            raise AlreadyHasDev(tag)
        item["dev"] = dev

    def set_builder_mount_point(self, tag, mount_point, cached=False):
        item = self._get(tag)
        if item["builder_mount_point"] is not None:
            raise AlreadyMounted(tag)
        item["builder_mount_point"] = mount_point
        item["cached"] = cached

    def set_fstype(self, tag, fstype):
        item = self._get(tag)
        if item["fstype"] is not None:
            raise AlreadyHasFsType(tag)
        item["fstype"] = fstype

    def set_target_mount_point(self, tag, target_mount_point):
        item = self._get(tag)
        if item["target_mount_point"] is not None:
            raise AlreadyHasTargetMountPoint(tag)
        item["target_mount_point"] = target_mount_point

    def get_builder_from_target_mount_point(self, target_mount_point):
        for item in self._tags.values():
            if item["target_mount_point"] == target_mount_point:
                return item["builder_mount_point"]
        raise NeedBothMountPoints(target_mount_point)

    def _get(self, tag):
        item = self._tags.get(tag)
        if item is None:
            raise UnknownTag(tag)
        return item


class TagInUse(Exception):
    def __init__(self, tag):
        super().__init__("Tag already used: {}".format(tag))


class UnknownTag(Exception):
    def __init__(self, tag):
        super().__init__("Unknown tag: {}".format(tag))


class AlreadyHasDev(Exception):
    def __init__(self, tag):
        super().__init__("Already has device: {}".format(tag))


class AlreadyMounted(Exception):
    def __init__(self, tag):
        super().__init__("Already mounted tag: {}".format(tag))


class AlreadyHasFsType(Exception):
    def __init__(self, tag):
        super().__init__("Already has filesytem type: {}".format(tag))


class AlreadyHasTargetMountPoint(Exception):
    def __init__(self, tag):
        super().__init__("Already has target mount point: {}".format(tag))


class NeedBothMountPoints(Exception):
    def __init__(self, target_mp):
        super().__init__("Need both mount points set, target: {}".format(target_mp))



import subprocess

def _runcmd(ctx, argv):
    p = subprocess.Popen(argv, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = p.communicate("")
    ctx['stdout'] = stdout
    ctx['stderr'] = stderr
    ctx['exit'] = p.returncode

def _binary(basename):
    return os.path.join(srcdir, basename)

def given_file(ctx, filename=None):
    with open(filename, 'wb') as f:
        f.write(get_file(filename))
 
def run_vmdb2(ctx, filename=None, output=None):
    vmdb2 = _binary('vmdb2')
    _runcmd(ctx, [vmdb2, filename, '-v', '--output', output])

def exit_code_is(ctx, exit_code=None):
    assert_eq(ctx['exit'], int(exit_code))

def stdout_contains(ctx, pat1=None, pat2=None):
    stdout = ctx.get('stdout', b'').decode('utf-8')
    i = stdout.find(pat1)
    assert i >= 0, "pat1 not found"
    i = stdout[i:].find(pat2)
    assert i >= 0, "pat2 not found after pat1"

def stdout_does_not_contain(ctx, pat1=None):
    stdout = ctx.get('stdout', b'').decode('utf-8')
    i = stdout.find(pat1)
    assert i == -1, "pattern found"



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import unittest

import vmdb


class StepRunnerListTests(unittest.TestCase):
    def test_is_empty_initially(self):
        steps = vmdb.StepRunnerList()
        self.assertEqual(len(steps), 0)

    def test_adds_a_runner(self):
        steps = vmdb.StepRunnerList()
        runner = DummyStepRunner()
        steps.add(runner)
        self.assertEqual(len(steps), 1)

    def test_finds_correct_runner(self):
        steps = vmdb.StepRunnerList()
        keyspec = {"foo": str, "bar": str}
        runner = DummyStepRunner(keyspec=keyspec)
        steps.add(runner)
        found = steps.find({"foo": "foo", "bar": "bar"})
        self.assertEqual(runner, found)

    def test_raises_error_if_runner_not_found(self):
        steps = vmdb.StepRunnerList()
        keyspec = {"foo": str, "bar": str}
        runner = DummyStepRunner(keyspec=keyspec)
        steps.add(runner)
        with self.assertRaises(vmdb.NoMatchingRunner):
            steps.find({"foo": "foo"})

    def test_raises_error_if_wrong_step_key_values(self):
        steps = vmdb.StepRunnerList()
        keyspec = {"foo": str}
        runner = DummyStepRunner(keyspec=keyspec)
        steps.add(runner)
        with self.assertRaises(vmdb.StepKeyWrongValueType):
            steps.find({"foo": 42})


class DummyStepRunner(vmdb.StepRunnerInterface):
    def __init__(self, keyspec=None):
        self.keyspec = keyspec

    def get_key_spec(self):
        return self.keyspec

    def run(self, *args):
        pass


class StepRunnerGetKeyValuesTests(unittest.TestCase):
    def test_returns_values_from_step_for_mandatory_keys(self):
        keyspec = {"foo": str}
        runner = DummyStepRunner(keyspec=keyspec)
        self.assertEqual(runner.get_values({"foo": "bar"}), {"foo": "bar"})

    def test_raises_error_for_missing_mandatory_key(self):
        keyspec = {"foo": str}
        runner = DummyStepRunner(keyspec=keyspec)
        with self.assertRaises(vmdb.StepKeyMissing):
            runner.get_values({})

    def test_raises_error_for_wrong_type_of_value_for_mandatory_key(self):
        keyspec = {"foo": str}
        runner = DummyStepRunner(keyspec=keyspec)
        with self.assertRaises(vmdb.StepKeyWrongValueType):
            runner.get_values({"foo": 42})

    def test_returns_default_value_for_missing_optional_key(self):
        keyspec = {"foo": "bar"}
        runner = DummyStepRunner(keyspec=keyspec)
        self.assertEqual(runner.get_values({}), {"foo": "bar"})

    def test_returns_actual_value_for_optional_key(self):
        keyspec = {"foo": "bar"}
        runner = DummyStepRunner(keyspec=keyspec)
        self.assertEqual(runner.get_values({"foo": "yo"}), {"foo": "yo"})

    def test_raises_error_for_wrong_type_of_value_for_optional_key(self):
        keyspec = {"foo": "bar"}
        runner = DummyStepRunner(keyspec=keyspec)
        with self.assertRaises(vmdb.StepKeyWrongValueType):
            runner.get_values({"foo": 42})



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import os

import vmdb


class AptPlugin(vmdb.Plugin):
    def enable(self):
        self.app.step_runners.add(AptStepRunner())


class AptStepRunner(vmdb.StepRunnerInterface):
    def get_key_spec(self):
        return {"apt": str, "packages": [], "tag": "", "fs-tag": "", "clean": True, "recommends": True}

    def run(self, values, settings, state):
        operation = values["apt"]
        if operation != "install":
            raise Exception('"apt" must always have value "install"')

        packages = values["packages"]
        recommends = values["recommends"]
        tag = values.get("tag") or None
        if tag is None:
            tag = values["fs-tag"]
        mount_point = state.tags.get_builder_mount_point(tag)

        if not self.got_eatmydata(state):
            self.install_packages(mount_point, [], recommends, ["eatmydata"])
            state.got_eatmydata = True
        self.install_packages(mount_point, ["eatmydata"], recommends, packages)

        if values["clean"]:
            self.clean_cache(mount_point)

    def got_eatmydata(self, state):
        return hasattr(state, "got_eatmydata") and getattr(state, "got_eatmydata")

    def install_packages(self, mount_point, argv_prefix, recommends, packages):
        env = os.environ.copy()
        env["DEBIAN_FRONTEND"] = "noninteractive"

        vmdb.runcmd_chroot(mount_point, argv_prefix + ["apt-get", "update"], env=env)

        rec = ''
        if not recommends:
            rec = '--no-install-recommends'

        vmdb.runcmd_chroot(
            mount_point,
            argv_prefix + ["apt-get", "-y", "--no-show-progress", rec, "install"] + packages,
            env=env,
        )

    def clean_cache(self, mount_point):
        env = os.environ.copy()
        env["DEBIAN_FRONTEND"] = "noninteractive"

        vmdb.runcmd_chroot(mount_point, ["apt-get", "clean"], env=env)



# Copyright 2018  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import unittest


import vmdb


class TagsTests(unittest.TestCase):
    def test_lists_no_tags_initially(self):
        tags = vmdb.Tags()
        self.assertEqual(tags.get_tags(), [])

    def test_tells_if_tag_exists(self):
        tags = vmdb.Tags()
        self.assertFalse(tags.has_tag("foo"))
        tags.append("foo")
        self.assertTrue(tags.has_tag("foo"))
        self.assertEqual(tags.get_tags(), ["foo"])

    def test_remembers_order(self):
        tags = vmdb.Tags()
        tags.append("foo")
        tags.append("bar")
        self.assertTrue(tags.get_tags(), ["foo", "bar"])

    def test_get_dev_raises_error_for_unknown_tag(self):
        tags = vmdb.Tags()
        with self.assertRaises(vmdb.UnknownTag):
            tags.get_dev("does-not-exist")

    def test_getting_builder_mount_point_raises_error_for_unknown_tag(self):
        tags = vmdb.Tags()
        with self.assertRaises(vmdb.UnknownTag):
            tags.get_builder_mount_point("does-not-exist")

    def test_raises_error_for_reused_tag(self):
        tags = vmdb.Tags()
        tags.append("tag")
        with self.assertRaises(vmdb.TagInUse):
            tags.append("tag")

    def test_sets_dev(self):
        tags = vmdb.Tags()
        tags.append("first")
        tags.set_dev("first", "/dev/foo")
        self.assertEqual(tags.get_tags(), ["first"])
        self.assertEqual(tags.get_dev("first"), "/dev/foo")
        self.assertEqual(tags.get_builder_mount_point("first"), None)

    def test_adds_builder_mount_point(self):
        tags = vmdb.Tags()
        tags.append("first")
        tags.set_builder_mount_point("first", "/mnt/foo")
        self.assertEqual(tags.get_tags(), ["first"])
        self.assertEqual(tags.get_dev("first"), None)
        self.assertEqual(tags.get_builder_mount_point("first"), "/mnt/foo")

    def test_builder_mount_point_is_uncached_by_default(self):
        tags = vmdb.Tags()
        tags.append("first")
        tags.set_builder_mount_point("first", "/mnt/foo")
        self.assertFalse(tags.is_cached("first"))

    def test_builder_mount_point_can_be_made_cached(self):
        tags = vmdb.Tags()
        tags.append("first")
        tags.set_builder_mount_point("first", "/mnt/foo", cached=True)
        self.assertTrue(tags.is_cached("first"))

    def test_set_dev_raises_error_for_unknown_tag(self):
        tags = vmdb.Tags()
        with self.assertRaises(vmdb.UnknownTag):
            tags.set_dev("first", "/mnt/foo")

    def test_set_builder_mount_point_raises_error_for_unknown_tag(self):
        tags = vmdb.Tags()
        with self.assertRaises(vmdb.UnknownTag):
            tags.set_builder_mount_point("first", "/mnt/foo")

    def test_set_builder_mount_point_raises_error_for_double_mount(self):
        tags = vmdb.Tags()
        tags.append("first")
        tags.set_builder_mount_point("first", "/mnt/foo")
        with self.assertRaises(vmdb.AlreadyMounted):
            tags.set_builder_mount_point("first", "/mnt/foo")

    def test_set_dev_raises_error_for_double_dev(self):
        tags = vmdb.Tags()
        tags.append("first")
        tags.set_dev("first", "/dev/foo")
        with self.assertRaises(vmdb.AlreadyHasDev):
            tags.set_dev("first", "/dev/foo")

    def test_set_fstype(self):
        tags = vmdb.Tags()
        tags.append("first")
        tags.set_fstype("first", "ext4")
        self.assertEqual(tags.get_fstype("first"), "ext4")

    def test_set_fstype_raises_error_for_double_fstype(self):
        tags = vmdb.Tags()
        tags.append("first")
        tags.set_fstype("first", "ext3")
        with self.assertRaises(vmdb.AlreadyHasFsType):
            tags.set_fstype("first", "ext4")

    def test_set_target_mount_point(self):
        tags = vmdb.Tags()
        tags.append("first")
        tags.set_target_mount_point("first", "/boot")
        self.assertEqual(tags.get_target_mount_point("first"), "/boot")

    def test_set_target_mount_point_raises_error_for_double_target_mount_point(self):
        tags = vmdb.Tags()
        tags.append("first")
        tags.set_target_mount_point("first", "/boot")
        with self.assertRaises(vmdb.AlreadyHasTargetMountPoint):
            tags.set_target_mount_point("first", "/")

    def test_raises_error_if_both_mount_points_not_set(self):
        tags = vmdb.Tags()
        tags.append("first")
        tags.set_target_mount_point("first", "/boot")
        with self.assertRaises(vmdb.NeedBothMountPoints):
            tags.get_builder_from_target_mount_point("/")

    def test_returns_builder_when_given_target_mount_point(self):
        tags = vmdb.Tags()
        tags.append("first")
        tags.set_builder_mount_point("first", "/mnt/foo")
        tags.set_target_mount_point("first", "/boot")
        self.assertEqual(tags.get_builder_from_target_mount_point("/boot"), "/mnt/foo")



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import vmdb


class MkfsPlugin(vmdb.Plugin):
    def enable(self):
        self.app.step_runners.add(MkfsStepRunner())


class MkfsStepRunner(vmdb.StepRunnerInterface):
    def get_key_spec(self):
        return {"mkfs": str, "partition": str, "label": "", "options": ""}

    def run(self, values, settings, state):
        fstype = values["mkfs"]
        tag = values["partition"]
        device = state.tags.get_dev(tag)

        if not isinstance(fstype, str):
            raise vmdb.NotString("mkfs", fstype)
        if not isinstance(tag, str):
            raise vmdb.NotString("mkfs: tag", tag)
        if not isinstance(device, str):
            raise vmdb.NotString("mkfs: device (for tag)", device)

        cmd = ["/sbin/mkfs", "-t", fstype]
        label = values["label"] or None
        if label:
            if fstype == "vfat":
                cmd.append("-n")
            elif fstype == "f2fs":
                cmd.append("-l")
            else:
                cmd.append("-L")
            cmd.append(label)

        options = values["options"] or None
        if options:
            for opt in options.split(' '):
                cmd.append(opt)

        cmd.append(device)
        vmdb.runcmd(cmd)

        state.tags.set_fstype(tag, fstype)



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import os

import vmdb


class ShellPlugin(vmdb.Plugin):
    def enable(self):
        self.app.step_runners.add(ShellStepRunner())


class ShellStepRunner(vmdb.StepRunnerInterface):
    def get_key_spec(self):
        return {"shell": str, "root-fs": str}

    def run(self, step, settings, state):
        shell = step["shell"]
        fs_tag = step["root-fs"]

        env = dict(os.environ)
        env["ROOT"] = state.tags.get_builder_mount_point(fs_tag)
        vmdb.runcmd(["sh", "-ec", shell], env=env)



# Copyright 2018  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import os

import vmdb


class LvcreatePlugin(vmdb.Plugin):
    def enable(self):
        self.app.step_runners.add(LvcreateStepRunner())


class LvcreateStepRunner(vmdb.StepRunnerInterface):
    def get_key_spec(self):
        return {"lvcreate": str, "name": str, "size": str}

    def run(self, values, settings, state):
        vgname = values["lvcreate"]
        lvname = values["name"]
        size = values["size"]

        vmdb.runcmd(["lvcreate", "-qq", "--name", lvname, "--size", size, vgname])

        lvdev = "/dev/{}/{}".format(vgname, lvname)
        assert os.path.exists(lvdev)
        state.tags.append(lvname)
        state.tags.set_dev(lvname, lvdev)



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


# Installing GRUB onto a disk image is a bit of a black art. I haven't
# found any good documentation for it. This plugin is written based on
# de-ciphering the build_openstack_image script. Here is an explanation
# of what I _THINK_ is happening.
#
# The crucial command is grub-install. It needs a ton of options to
# work correctly: see below in the code for the list, and the manpage
# for an explanation of what each of them means. We will be running
# grub-install in a chroot so that we use the version in the Debian
# version we're installing, rather than the host system, which might
# be any Debian version.
#
# To run grub-install in a chroot, we need to set up the chroot in
# various ways. Firstly, we need to tell grub-install which device
# file the image has. We can't just give it the image file itself,
# since it isn't inside the chroot, so instead we arrange to have a
# loop block device that covers the whole image file, and we bind
# mount /dev into the chroot so the device is available.
#
# grub-install seems to also require /proc and /sys so we bind mount
# /sys into the chroot as well. /proc is already mounted otherwise.
#
# We install the UEFI version of GRUB, and for that we additionally
# bind mount the EFI partition in the image. Oh yeah, you MUST have
# one.
#
# We also make sure the right GRUB package is installed in the chroot,
# before we run grub-install.
#
# Further, there's some configuration tweaking we need to do. See the
# code. Don't ask me why they're necessary.
#
# For cleanliness, we also undo any bind mounts into the chroot. Don't
# want to leave them in case they cause trouble.
#
# Note that this is currently assuming that UEFI and either the amd64
# (a.k.a. x86_64) or arm64 (a.k.a. aarch64) architectures are being
# used. These should probably not be hardcoded. Patch welcome.

# To use this plugin: write steps to create a root filesystem, and an
# VFAT filesystem to be mounted as /boot/efi. Install Debian onto the
# root filesystem. Then install grub with a step like this:
#
#         - grub: uefi
#           tag: root-part
#           efi: efi-part
#
# Here: "tag" is the tag for the root filesystem (and corresponding
# partition), and efi is tag for the EFI partition.
#
# The grub step will take of the rest.


import logging
import os
import re

import vmdb


class GrubPlugin(vmdb.Plugin):
    def enable(self):
        self.app.step_runners.add(GrubStepRunner())


class GrubStepRunner(vmdb.StepRunnerInterface):
    def get_key_spec(self):
        return {
            "grub": str,
            "root-fs": "",
            "efi": "",
            "efi-part": "",
            "prep": "",
            "console": "",
            "tag": "",
            "image-dev": "",
            "quiet": False,
            "timeout": 0,
        }

    def run(self, values, settings, state):
        state.grub_mounts = []
        flavor = values["grub"]
        if flavor == "uefi":
            self.install_uefi(values, settings, state)
        elif flavor == "bios":
            self.install_bios(values, settings, state)
        elif flavor == "ieee1275":
            self.install_ieee1275(values, settings, state)
        else:
            raise Exception("Unknown GRUB flavor {}".format(flavor))

    def grub_uefi_variant(self, state):
        variants = {
            "amd64": ("grub-efi-amd64", "x86_64-efi"),
            "i386": ("grub-efi-ia32", "i386-efi"),
            "arm64": ("grub-efi-arm64", "arm64-efi"),
            "armhf": ("grub-efi-arm", "arm-efi"),
        }
        try:
            return variants[state.arch]
        except KeyError:
            raise Exception(
                'GRUB UEFI package and target for "{}" unknown'.format(state.arch)
            )

    def install_uefi(self, values, settings, state):
        efi = values["efi"] or None
        efi_part = values["efi-part"] or None
        if efi is None and efi_part is None:
            raise Exception('"efi" or "efi-part" required in UEFI GRUB installation')

        vmdb.progress("Installing GRUB for UEFI")
        (grub_package, grub_target) = self.grub_uefi_variant(state)
        self.install_grub(values, settings, state, grub_package, grub_target)

    def install_bios(self, values, settings, state):
        vmdb.progress("Installing GRUB for BIOS")
        grub_package = "grub-pc"
        grub_target = "i386-pc"
        self.install_grub(values, settings, state, grub_package, grub_target)

    def install_ieee1275(self, values, settings, state):
        vmdb.progress("Installing GRUB for IEEE1275")
        grub_package = "grub-ieee1275"
        grub_target = "powerpc-ieee1275"
        self.install_grub(values, settings, state, grub_package, grub_target)

    def install_grub(self, values, settings, state, grub_package, grub_target):
        console = values["console"] or None

        tag = values["tag"] or values["root-fs"] or None
        root_dev = state.tags.get_dev(tag)
        chroot = state.tags.get_builder_mount_point(tag)

        image_dev = values["image-dev"] or None
        if image_dev is None:
            image_dev = self.get_image_loop_device(root_dev)

        efi = values["efi"] or None
        efi_part = values["efi-part"] or None
        if efi is not None:
            efi_dev = state.tags.get_dev(efi)
        elif efi_part is not None:
            efi_dev = state.tags.get_dev(efi_part)
        else:
            efi_dev = None

        prep = values["prep"] or None
        if prep:
            prep_dev = state.tags.get_dev(prep)
        else:
            prep_dev = None

        quiet = values["quiet"]

        self.bind_mount_many(chroot, ["/dev", "/sys", "/proc"], state)
        if efi_dev:
            pn = efi_dev[-1]
            vmdb.runcmd(["parted", "-s", image_dev, "set", pn, "esp", "on" ])
            self.mount(chroot, efi_dev, "/boot/efi", state)
        elif prep_dev:
            pn = prep_dev[-1]
            vmdb.runcmd(["parted", "-s", image_dev, "set", pn, "prep", "on" ])
            image_dev = prep_dev
        self.install_package(chroot, grub_package)

        kernel_params = [
            "biosdevname=0",
            "net.ifnames=0",
            "consoleblank=0",
            "systemd.show_status=true",
            "rw",
        ]
        if console == "serial":
            if 'ppc64' in state.arch:
                kernel_params.extend(
                    ["loglevel=3", "console=tty0", "console=hvc0,115200n8"]
                )
            elif 'arm' in state.arch:
                kernel_params.extend(
                    ["loglevel=3", "console=tty0", "console=ttyAMA0,115200n8"]
                )
            else:
                kernel_params.extend(
                    ["loglevel=3", "console=tty0", "console=ttyS0,115200n8"]
                )

        if quiet:
            kernel_params.extend(
                [
                    "quiet",
                    "systemd.show_status=false",
                    "rd.systemd.show_status=false",
                    "systemd.show_status=false",
                ]
            )

        self.set_grub_cmdline_config(chroot, kernel_params)
        self.add_grub_crypto_disk(chroot)
        self.set_grub_timeout(chroot, values["timeout"])
        if console == "serial":
            self.add_grub_serial_console(chroot)

        vmdb.runcmd_chroot(chroot, ["grub-mkconfig", "-o", "/boot/grub/grub.cfg"])
        vmdb.runcmd_chroot(
            chroot,
            [
                "grub-install",
                "--target=" + grub_target,
                "--no-nvram",
                "--force-extra-removable",
                "--no-floppy",
                "--modules=part_msdos part_gpt",
                "--grub-mkdevicemap=/boot/grub/device.map",
                image_dev,
            ],
        )

    #        self.unmount(state)

    def teardown(self, values, settings, state):
        self.unmount(state)

    def unmount(self, state):
        mounts = getattr(state, "grub_mounts", [])
        mounts.reverse()
        while mounts:
            mount_point = mounts.pop()
            try:
                vmdb.unmount(mount_point)
            except vmdb.NotMounted as e:
                logging.warning(str(e))

    def get_image_loop_device(self, partition_device):
        # We get /dev/mappers/loopXpY and return /dev/loopX
        # assert partition_device.startswith('/dev/mapper/loop')

        m = re.match(r"^/dev/mapper/(?P<loop>.*)p\d+$", partition_device)
        if m is None:
            raise Exception(
                "Do not understand partition device name {}".format(partition_device)
            )
        assert m is not None

        loop = m.group("loop")
        return "/dev/{}".format(loop)

    def bind_mount_many(self, chroot, paths, state):
        for path in paths:
            self.mount(chroot, path, path, state, mount_opts=["--bind"])

    def mount(self, chroot, path, mount_point, state, mount_opts=None):
        chroot_path = self.chroot_path(chroot, mount_point)
        if not os.path.exists(chroot_path):
            os.makedirs(chroot_path)

        if mount_opts is None:
            mount_opts = []

        vmdb.runcmd(["mount"] + mount_opts + [path, chroot_path])
        state.grub_mounts.append(chroot_path)

    def chroot_path(self, chroot, path):
        return os.path.normpath(os.path.join(chroot, "." + path))

    def install_package(self, chroot, package):
        env = os.environ.copy()
        env["DEBIAN_FRONTEND"] = "noninteractive"
        vmdb.runcmd_chroot(
            chroot, ["apt-get", "-y", "--no-show-progress", "install", package], env=env
        )

    def set_grub_cmdline_config(self, chroot, kernel_params):
        param_string = " ".join(kernel_params)

        filename = self.chroot_path(chroot, "/etc/default/grub")

        with open(filename) as f:
            text = f.read()

        lines = text.splitlines()
        lines = [
            line for line in lines if not line.startswith("GRUB_CMDLINE_LINUX_DEFAULT")
        ]
        lines.append('GRUB_CMDLINE_LINUX_DEFAULT="{}"'.format(param_string))

        with open(filename, "w") as f:
            f.write("\n".join(lines) + "\n")

    def add_grub_serial_console(self, chroot):
        filename = self.chroot_path(chroot, "/etc/default/grub")

        with open(filename, "a") as f:
            f.write("GRUB_TERMINAL=serial\n")
            f.write(
                'GRUB_SERIAL_COMMAND="serial --speed=115200 --unit=0 '
                '--word=8 --parity=no --stop=1"\n'
            )

    def add_grub_crypto_disk(self, chroot):
        filename = self.chroot_path(chroot, "/etc/default/grub")
        with open(filename, "a") as f:
            f.write("GRUB_ENABLE_CRYPTODISK=y\n")

    def set_grub_timeout(self, chroot, timeout):
        filename = self.chroot_path(chroot, "/etc/default/grub")
        with open(filename, "a") as f:
            f.write("GRUB_TIMEOUT={}\n".format(timeout))



#!/usr/bin/env python3
# Copyright (C) 2017  Lars Wirzenius <liw@liw.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from distutils.core import setup
import glob


import vmdb


setup(
    name="vmdb2",
    version=vmdb.__version__,
    description="create disk image with Debian installed",
    author="Lars Wirzenius",
    author_email="liw@liw.fi",
    url="http://liw.fi/vmdebootstrap/",
    scripts=["vmdb2"],
    packages=["vmdb", "vmdb.plugins"],
    data_files=[("share/man/man1", glob.glob("*.1"))],
)
